#ifndef _BMP_H_
#define _BMP_H_

#include <cstdio>
#include <cstdlib>

struct BMP_InfoHeader {
    unsigned int        biSize;                 /* size of the BMP_InfoHeader struct */
    unsigned int        biWidth;                /* bitmap width in pixels */
    unsigned int        biHeight;               /* bitmap height in pixels */
    unsigned short      biPlanes;               /* number of bitmap planes */
    unsigned short      biBitCount;             /* number of bits per pixel */
    unsigned int        biCompression;          /* compression format (0 == uncompressed) */
    unsigned int        biSizeImage;            /* size of image in bytes */
    unsigned int        biXPelsPerMeter;        /* pixels per meter in x direction */
    unsigned int        biYPelsPerMeter;        /* pixels per meter in y direction */
    unsigned int        biClrUsed;              /* number of colors used */
    unsigned int        biClrImportant;         /* number of important colors used */
};

struct BMP_FileHeader {
    unsigned short        bfType;                /* file type; must be "BM" */
    unsigned short        bfSize;                /* length of the file;
                                                    sizeof(BMP_FileHeader) +
                                                    sizeof(BMP_InfoHeader) +
                                                    size of color table +
                                                    size of DIB bits */
    unsigned short        bfFill1;              /* remainder of the bfSize field */
    unsigned short        bfReserved1;          /* don't care (set to 0)*/
    unsigned short        bfReserved2;          /* don't care (set to 0)*/
    unsigned short        bfOffBits;            /* offset from beginning of file */
    unsigned short        bfFill2;              /* remainder of the bfOffBits field */
};

#define BITMAP_ID   ('B' + ('M'<<8))

unsigned char *loadBMP(const char *filename, int &w, int &h)
{
    FILE *filePtr;                              // the file pointer
    BMP_FileHeader      bitmapFileHeader;       // bitmap file header
    BMP_InfoHeader      bitmapInfoHeader;
    unsigned char     * bitmapImage;            // bitmap image data

    // open filename in "read binary" mode
    filePtr = fopen(filename, "rb");
    if (filePtr == 0)
        return 0;

    // read the bitmap file header
    fread(&bitmapFileHeader, sizeof(BMP_FileHeader), 1, filePtr);

    // verify that this is a bitmap by checking for the universal bitmap id
    if (bitmapFileHeader.bfType != BITMAP_ID) {
        fclose(filePtr);
        return 0;
    }

    // read the bitmap information header
    fread(&bitmapInfoHeader, sizeof(BMP_InfoHeader), 1, filePtr);
    w = bitmapInfoHeader.biWidth, h = bitmapInfoHeader.biHeight;

    // move file pointer to beginning of bitmap data
    fseek(filePtr, bitmapFileHeader.bfOffBits, SEEK_SET);
    if(bitmapInfoHeader.biSizeImage == 0)
        bitmapInfoHeader.biSizeImage = w * h * bitmapInfoHeader.biBitCount/8;
    // allocate enough memory for the bitmap image data
    bitmapImage = (unsigned char*)malloc(bitmapInfoHeader.biSizeImage);

    // verify memory allocation
    if (!bitmapImage) {
        free(bitmapImage);
        fclose(filePtr);
        return 0;
    }

    // read in the bitmap image data
    fread(bitmapImage, bitmapInfoHeader.biSizeImage, 1, filePtr);
    for(int i=0; i<bitmapInfoHeader.biSizeImage/3-1; i++){
        unsigned char r = bitmapImage[i*3+0];
        bitmapImage[i*3+0] = bitmapImage[i*3+2];
        bitmapImage[i*3+2] = r;
    }

    // close the file and return the bitmap image data
    fclose(filePtr);
    
    return bitmapImage;
}

#endif //_BMP_H_
